/**
  *  \file arch/win32-bcc55/dirs.cc
  *  \brief Filename Parsing and Directory Handling for Windows
  */
#include <list>
#include <cctype>
#include <iostream>
#include <cerrno>
#include <cstring>
#include <vector>
#include <windows.h>
#include "io/dirs.h"
#include "cpluslib/assert.h"
#include "util/except.h"

namespace {
    struct Init {
        Init();
    };
}

static Init initializer;

/** Initialize Win32 platform stuff. */
Init::Init()
{
    /* Prevent Windows from displaying a dialog box when we access a
       disk drive which is not ready. */
    SetErrorMode(SEM_FAILCRITICALERRORS);
}

static void
frobSlash(string_t& name)
{
    string_t::size_type n = 0;
    while ((n = name.find('/', n)) != name.npos)
        name[n] = '\\';
}

/** Test whether file name is absolute.
    \returns true iff file name is absolute, that is, it has
    the same meaning on this computer no matter what the current
    working directory is. */
bool
Directory::isAbsoluteName(const string_t name)
{
    // FIXME: we don't support UNC names
    return (name.length() > 2 && name[1] == ':'
            && (name[2] == '\\' || name[2] == '/'));
}

/** Test whether character is path separator.
    \returns true iff c is a path separator char (that is, something
    that can not occur in a file "basename") */
bool
Directory::isSeparatorChar(char c)
{
    return c == '\\' || c == ':' || c == '/';
}

/** Make file name from two components.
    \param dir     working directory
    \param name    file name
    \returns file name that names the file that would be accessed when
    we were in directory /dir/ and opened /name/. In particular, if
    /name/ is absolute, /dir/ is ignored. The returned file name is
    canonical (not necessarily absolute). */
string_t
Directory::makeFileName(string_t dir, const string_t name)
{
    if (isAbsoluteName(name))
        return getCanonicalName(name);
    if (name.length() >= 2 && name[1] == ':')
        return getCanonicalName(name);
    if (dir.length() && !isSeparatorChar(dir[dir.length()-1]))
        dir += PATH_SEPARATOR;
    return getCanonicalName(dir + name);
}

/** Canonicalize directory name. Generates the canonical name for the
    specified file name/directory. It does so by removing redundant
    components (e.g. in Unix notation, `foo/./bar/../x' is the same
    as `foo/x').

    getCanonicalName("") returns the notation for the "current
    directory". */
string_t
Directory::getCanonicalName(string_t dir)
{
    frobSlash(dir);

    /* If there is a drive letter, remove it and save it for later. */
    string_t prefix;
    if (dir.length() >= 1 && dir[1] == ':') {
        prefix.assign(dir, 0, 2);
        dir.erase(0, 2);
    }

    if (!dir.length())
        return prefix + ".";

    bool is_absolute;
    int  up_count = 0;
    std::list<string_t> out;
    if (dir[0] == PATH_SEPARATOR) {
        is_absolute = true;
        dir.erase(0, 1);
    } else {
        is_absolute = false;
    }

    string_t::size_type pos = 0;
    while (pos < dir.length()) {
        string_t::size_type next = dir.find(PATH_SEPARATOR, pos);
        if (next == string_t::npos)
            next = dir.length();
        if (next == pos) {
            /* ignore consecutive slashes */
        } else if (next == pos+1 && dir[pos] == '.') {
            /* ignore "." components */
        } else if (next == pos+2 && dir.compare(pos, 2, "..", 2) == 0) {
            /* ".." component */
            if (!out.empty())
                out.pop_back();
            else
                ++up_count;
        } else {
            out.push_back(string_t(dir, pos, next-pos));
        }
        if (next >= dir.length())
            break;
        pos = next+1;
    }

    /* now generate output */
    string_t s;
    if (is_absolute) {
        s += PATH_SEPARATOR;
    } else {
        while (up_count--) {
            s += "..";
            s += PATH_SEPARATOR;
        }
    }
    for (std::list<string_t>::iterator i = out.begin(); i != out.end(); ++i) {
        s += *i;
        s += PATH_SEPARATOR;
    }
    if (s.length() > 1)
        s.erase(s.length() - 1);
    else if (!s.length())
        return prefix + ".";
    return prefix + s;
}

/** Get absolute name. Returns the canonical, absolute name for the
    specified directory.
    \post isAbsoluteName(result) */
string_t
Directory::getAbsoluteName(string_t dir)
{
    using namespace std;
    frobSlash(dir);

    // we used to do this by hand, but Win32 has a handy function to do it:
    LPTSTR tmpp;
    char tmpc;
    int bytes_needed = GetFullPathName(dir.c_str(), 0, &tmpc, &tmpp);
    if (bytes_needed < 2) {
        // Failure, do the best we can do
        return getCanonicalName(dir);
    } else {
        // okay, use Win32 function to do it:
        std::vector<char> buffer(bytes_needed+1);
        GetFullPathName(dir.c_str(), bytes_needed, &buffer.front(), &tmpp);
        return string_t(&buffer.front());
    }
}

/** Get file name part ("basename") of a name.
    This can be used to split a file name into pieces.
    \see getDirName
    \inv getAbsoluteName(dir) == getAbsoluteName(makeFileName(getDirName(dir), getFileName(dir))) */
string_t
Directory::getFileName(string_t dir)
{
    string_t::size_type d = dir.find_last_of("\\:/");
    if (d != string_t::npos && (dir[d] == '\\' || dir[d] == '/' || d == 1))
        return dir.substr(d+1);
    else
        return dir;
}

/** Get directory name part of a name. \see getFileName */
string_t
Directory::getDirName(string_t dir)
{
    frobSlash(dir);
    string_t::size_type d = dir.find_last_of("\\:");
    if (d == 0 && dir[d] == '\\')
        /* `\file' */
        return string_t(1, PATH_SEPARATOR);
    else if (d == 1 && dir[1] == ':')
        /* `a:file' */
        return dir.substr(0, 2);
    else if (d == 2 && dir[1] == ':' && dir[2] == '\\')
        /* `a:\file' */
        return dir.substr(0, 3);
    else if (d != string_t::npos && dir[d] == '\\')
        /* `dir\file' */
        return dir.substr(0, d);
    else
        /* `file' */
        return ".";
}

/** Get current working directory.
    \post isAbsoluteName(result)
    \throws FileIOException on error */
string_t
Directory::getWorkingDirectory()
{
    std::size_t len = 0;
    char* buffer;
    static const std::size_t PATH_MAX = 256;

    while (1) {
        len += PATH_MAX;
        buffer = new char[len];
        if (GetCurrentDirectory(len, buffer) < len)
            break;
        delete[] buffer;
    }

    string_t result(buffer);
    delete[] buffer;
    return result;
}

/********************* Directory class **********************/

/* FIXME: This is mostly a copy of the Unix version.
   It should be rewritten to use the native Windows functions
   instead. */

namespace {
    class WindowsDirectory : public Directory {
        class Entry;
        class Enum;
        string_t dir;
        friend class Enum;
        friend class Entry;
     public:
        WindowsDirectory(string_t aname);
        ~WindowsDirectory();
        Ptr<DirectoryEntry> getEntryByName(string_t name);
        Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries();
        Ptr<Directory> getParent();
        string_t getName();
    };

    class WindowsDirectoryWithTitle : public WindowsDirectory {
        string_t title;
     public:
        WindowsDirectoryWithTitle(string_t name, string_t title)
            : WindowsDirectory(name), title(title) { }
        string_t getTitle()
            { return title; }
    };

    class WindowsRootEnum : public Enumeration<Ptr<Directory> > {
        char* buffer;
        char* current;
        void skipUnready();
     public:
        WindowsRootEnum();
        ~WindowsRootEnum();
        bool hasMoreElements();
        Ptr<Directory> getNextElement();
    };
}

Ptr<Directory>
Directory::openDir(string_t dir)
{
    return new WindowsDirectory(dir);
}

//////////////////////////////////////////////////////////////////////

/**
  *  \class WindowsDirectory
  *  \brief Directory on File System
  *
  *  This class represents an actual file system directory.
  */

/** File System Directory Entry. Implementation of DirectoryEntry
    for file system directories. */
class WindowsDirectory::Entry : public DirectoryEntry {
    friend class WindowsDirectory;
    friend class WindowsDirectory::Enum;
    Ptr<WindowsDirectory> parent;
    Ptr<Entry> link;
 protected:
    void getInfo();
 public:
    Entry(Ptr<WindowsDirectory> fsd, string_t aname);
    Ptr<Directory> openDir();
    Ptr<Stream>    openFile(int mode);
    Ptr<Directory> openParent();
    void erase();
    void renameTo(string_t name);
};

WindowsDirectory::Entry::Entry(Ptr<WindowsDirectory> fsd, string_t aname)
    : DirectoryEntry(aname), parent(fsd), link(0)
{ }

Ptr<Directory>
WindowsDirectory::Entry::openDir()
{
    return new WindowsDirectory(name);
}

Ptr<Stream>
WindowsDirectory::Entry::openFile(int mode)
{
    return Stream::openFile(name, mode);
}

Ptr<Directory>
WindowsDirectory::Entry::openParent()
{
    return parent;
}

void
WindowsDirectory::Entry::erase()
{
    if (!DeleteFile(name.c_str()))
        throw FileIOException(name, "Can't delete"); /* FIXME: GetLastError */
    clearInfo();
}

void
WindowsDirectory::Entry::renameTo(string_t newname)
{
    string_t newpath = Directory::makeFileName(parent->dir, newname);
    if (!MoveFile(name.c_str(), newname.c_str()))
        throw FileIOException(name, "Can't rename"); /* FIXME: GetLastError */
    clearInfo();
}

void
WindowsDirectory::Entry::getInfo()
{
    WIN32_FIND_DATA data;
    HANDLE h = FindFirstFile(name.c_str(), &data);
    if (h != INVALID_HANDLE_VALUE) {
        FileType ft = (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? FT_DIRECTORY : FT_FILE;
        setInfo(data.nFileSizeLow, ft);
        FindClose(h);
    }
}

/** File System Directory Enumerator. Implementation of the directory
    enumeration for WindowsDirectory. */
class WindowsDirectory::Enum : public Enumeration<Ptr<DirectoryEntry> >
{
    Ptr<WindowsDirectory> dir;           ///< The directory we're using
    Ptr<WindowsDirectory::Entry> entry;  ///< Current entry
    WIN32_FIND_DATA find_data;
    HANDLE find_handle;                  ///< Directory handle. INVALID_HANDLE_VALUE when done.
    void next(bool force);
 public:
    Enum(Ptr<WindowsDirectory> adir);
    ~Enum();
    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

WindowsDirectory::Enum::Enum(Ptr<WindowsDirectory> adir)
    : dir (adir),
      entry (0),
      find_handle (INVALID_HANDLE_VALUE)
{
    find_handle = FindFirstFile(makeFileName(dir->dir, "*").c_str(), &find_data);
    if (find_handle == INVALID_HANDLE_VALUE)
        throw FileOpenException(dir->dir, "Can't open directory");
    next(false);
}

WindowsDirectory::Enum::~Enum()
{
    if (find_handle != INVALID_HANDLE_VALUE)
        FindClose(find_handle);
}

/** Advance to next entry. Sets entry to the next directory entry,
    and skip entries we do not want.
    \param force true to advance an entry, false if find_data already
                 contains a valid entry */
void
WindowsDirectory::Enum::next(bool force)
{
    while (1) {
        if (find_handle != INVALID_HANDLE_VALUE) {
            if (force && !FindNextFile(find_handle, &find_data)) {
                // end of directory reached
                FindClose(find_handle);
                find_handle = INVALID_HANDLE_VALUE;
                entry = 0;
                break;
            } else {
                // found a matching file
                if (strcmp(find_data.cFileName, ".") == 0 || strcmp(find_data.cFileName, "..") == 0) {
                    // we do not want this file
                } else {
                    // accept
                    entry = new WindowsDirectory::Entry(dir, Directory::makeFileName(dir->dir, find_data.cFileName));
                    break;
                }
            }
            // in case we're doing another run, force read of new entry
            force = true;
        } else {
            // we already were at the end of the directory
            entry = 0;
            break;
        }
    }
}

bool
WindowsDirectory::Enum::hasMoreElements()
{
    return entry.ptr();
}

Ptr<DirectoryEntry>
WindowsDirectory::Enum::getNextElement()
{
    ASSERT(entry.ptr());
    Ptr<DirectoryEntry> ret = entry;
    next(true);
    return ret;
}

/** Constructor. Construct a WindowsDirectory for the
    directory specified by aname. */
WindowsDirectory::WindowsDirectory(string_t aname)
    : dir(getCanonicalName(aname))
{ }

WindowsDirectory::~WindowsDirectory()
{ }

Ptr<Directory>
WindowsDirectory::getParent()
{
    string_t my_abs = getAbsoluteName(dir);
    if (my_abs.length() <= 3)
        return 0;
    else
        return new WindowsDirectory(makeFileName(my_abs, ".."));
}

Ptr<DirectoryEntry>
WindowsDirectory::getEntryByName(string_t name)
{
    return new Entry(this, makeFileName(dir, name));
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
WindowsDirectory::getEntries()
{
    return new Enum(this);
}

string_t
WindowsDirectory::getName()
{
    return dir;
}

WindowsRootEnum::WindowsRootEnum()
{
    char dummy;
    DWORD count = GetLogicalDriveStrings(0, &dummy);
    if (count == 0) {
        buffer = current = new char[1];
        *current = 0;
    } else {
        buffer = current = new char[count+1];
        GetLogicalDriveStrings(count, buffer);
        skipUnready();
    }
}

WindowsRootEnum::~WindowsRootEnum()
{
    delete[] buffer;
}

bool
WindowsRootEnum::hasMoreElements()
{
    return *current != 0;
}

Ptr<Directory>
WindowsRootEnum::getNextElement()
{
    string_t name = current;
    current += name.length() + 1;
    skipUnready();
    return new WindowsDirectoryWithTitle(name, name);
}

/* FIXME: This is a HACK. User code can not determine whether a file open
   fails because the file does not exist, or because the disk is not
   ready. Hence, directory reading code would bombard the disk drive with
   requests. We sweep the problem under the carpet by not including disk
   drives which are not ready in directory listings.
   
   A much better solution would be to modify openDir/openFile to remember
   when a disk access failed due to an inaccessible disk, and let future
   requests fail silently for, say, a second. This would need a windows-
   specific file I/O class, though. Right now, we use the iostream-based
   one.

   The downside is that this does not allow users to open the game
   selector, put in a disk, and access the game on the disk. However,
   this is a rather infrequent event I think. Apart from that, other
   removable media like USB sticks do not have a drive letter at all
   while plugged out, so why should disks? */
void
WindowsRootEnum::skipUnready()
{
    DWORD a, b, c, d;
    while (*current && !GetDiskFreeSpace(current, &a, &b, &c, &d))
        current += std::strlen(current) + 1;
}

Ptr<Enumeration<Ptr<Directory> > >
Directory::getRoots()
{
    return new WindowsRootEnum();
}
