/**
  *  \file client/cscr-sel.cc
  *  \brief Object Selection Window
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "cpluslib/holder.h"

/************************* WObjectSelectionTile *************************/

/*! \class WObjectSelectionTile
    \brief Interior of "Select-a-$THING" window

    This one provides the main widget of a select-a-thing window:
    browsing with up/down, selection.

    Derived classes must override its draw() method to do something
    useful. */

/** Constructor.
    \param type Type we're selecting from
    \pre Type has objects */
WObjectSelectionTile::WObjectSelectionTile(GObjectType& type)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, 10)),
      type(type), current(type.getCurrent())
{
    ASSERT(type.getCount() != 0);
}

WObjectSelectionTile::WObjectSelectionTile(GObjectType& type, Signal1<int>& sig_update)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, 10)),
      type(type), current(type.getCurrent())
{
    ASSERT(type.getCount() != 0);
    conn = sig_update.add(this, &WObjectSelectionTile::updateId);
}

WObjectSelectionTile::~WObjectSelectionTile()
{ }

/** Select an object by Id number.
    \pre getType().isExistant(id) */
void
WObjectSelectionTile::setCurrent(int id)
{
    ASSERT(getType().isExistant(id));
    if (id != current) {
        current = id;
        update();
    }
}

void
WObjectSelectionTile::updateId(int id)
{
    if (id == current)
        update();
}

namespace {
    class WObjectSelectionWindow : public UIWindow {
        WObjectSelectionTile& tile;
        GObjectType& type;
     public:
        WObjectSelectionWindow(const string_t title, WObjectSelectionTile& tile, GObjectType& type)
            : UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY),
              tile(tile),
              type(type)
            { }

        bool handleCommand(const UICommand& cmd, UIWidget& sender);
    };
}

bool
WObjectSelectionWindow::handleCommand(const UICommand& cmd, UIWidget& sender)
{
    switch (cmd.code) {
     case cm_Prev:
        tile.select();
        tile.setCurrent(type.getPrev(tile.getCurrent(), false));
        return true;
     case cm_Next:
        tile.select();
        tile.setCurrent(type.getNext(tile.getCurrent(), false));
        return true;
     case cm_PrevMarked:
        tile.select();
        tile.setCurrent(type.getPrev(tile.getCurrent(), true));
        return true;
     case cm_NextMarked:
        tile.select();
        tile.setCurrent(type.getNext(tile.getCurrent(), true));
        return true;
     case cm_CS_ToggleSel: {
        UIWidget::select();
        GObject& g = type.getObject(tile.getCurrent());
        g.setIsMarked(!g.isMarked());
        type.getUniverse().doScreenUpdates();
        return true;
     }
     case cm_Enter:
        stop(tile.getCurrent());
        return true;
     case cm_Escape:
        stop(0);
        return true;
    }
    return UIWindow::handleCommand(cmd, sender);
}

/** Do "select-a-$THING" window. Constructs the window and runs it.
    \param inter  instance of WObjectSelectionTile specialised for displaying the desired object type
    \param title  window title
    \return selected object Id, or zero if user exited with ESC */
int
WObjectSelectionTile::doStandardDialog(const string_t title)
{
    static const UIKeyTranslator::Init init[] = {
        { SDLK_UP,           cm_Prev },
        { SDLK_DOWN,         cm_Next },
        { SDLK_PAGEUP,       cm_Prev },
        { SDLK_PAGEDOWN,     cm_Next },
        { ss_Ctrl+SDLK_UP,   cm_PrevMarked },
        { ss_Ctrl+SDLK_DOWN, cm_NextMarked },
        { ss_Ctrl+SDLK_PAGEUP,   cm_PrevMarked },
        { ss_Ctrl+SDLK_PAGEDOWN, cm_NextMarked },
        { vk_WheelUp,        cm_Prev },
        { vk_WheelDown,      cm_Next },
        { '.',               cm_CS_ToggleSel }
    };

    ObjectHolder holder;
    WObjectSelectionWindow win(title, *this, type);
    UIGroup h(UIHBoxLayout::instance5);
    win.setLayout(UIVBoxLayout::instance5);

    win.add(*this);
    win.add(h);
    win.add(holder.add(new UIKeyTranslator(init, countof(init))));
    h.add(holder.add(new UIButton(cm_Prev,  "\036",            '-')).addCommand());
    h.add(holder.add(new UIButton(cm_Next,  "\037",            '+')).addCommand());
    h.add(holder.add(new UIButton(cm_Enter, _("ENTER - View"), SDLK_RETURN)).addCommand());
    h.add(holder.add(new UISpacer()));
    h.add(holder.add(new UIButton(cm_Escape, _("ESC"),         SDLK_ESCAPE)).addCommand());
    h.add(holder.add(new UIButton(cm_Help,  "H",               'H')).addCommand());
    win.pack();

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    return rv;
}
