/**
  *  \file game/act-ammo.cc
  *  \brief Ammunition Building Action
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "game/action.h"
#include "game/spec.h"

/** Constructor.
    \param pay       who will pay for the ammo, must be a planet
    \param receiver  who will receive the ammo, can be any container
    \pre we are playing pay and receiver, and planet pay has a base */
GStarbaseBuildAmmoAction::GStarbaseBuildAmmoAction(GPlanetTransfer& pay,
                                                   GCargoContainer& receiver)
    : GCargoCostTransaction(pay), pay(pay), receiver(receiver)
{
    for (unsigned i = 0; i < countof(added); ++i)
        added[i] = 0;
}

GStarbaseBuildAmmoAction::~GStarbaseBuildAmmoAction()
{ }

/** Convert index into cargo type. This class uses indexes
    [min_index,max_index] to identify the ammo types. This function
    returns the corresponding cargo type. */
GCargoType
GStarbaseBuildAmmoAction::getCargoTypeFromIndex(int index)
{
    if (index > 0 && index <= NUM_TORPS) {
        return (GCargoType(el_Torps + index-1));
    } else if (index == max_index) {
        return el_Fighters;
    } else {
        ASSERT(0);
        return el_Neutronium; // silence compiler
    }
}

/** Check whether receiver can have the specified ammo type.
    \param index type index, [min_index,max_index]
    \return true iff receiver can carry that type */
bool
GStarbaseBuildAmmoAction::canHave(int index) const
{
    return receiver.canHaveCargo(getCargoTypeFromIndex(index));
}

/** Get current amount of specified ammo type.
    \param index type index, [min_index,max_index]
    \return amount of specified type on target */
int32
GStarbaseBuildAmmoAction::getAmount(int index) const
{
    return receiver.getCargo(getCargoTypeFromIndex(index));
}

/** Set current amount on receiver.
    \param index type index, [min_index,max_index]
    \param amount new amount

    The new amount can be any value. If it is too small, or if the
    receiver cannot carry that much, it is adjusted. If it exceeds the
    payers resources, we will go into an error state until the amount
    is reduced again. */
void
GStarbaseBuildAmmoAction::setAmount(int index, int32 amount)
{
    // difference to total amount
    int32 diff = amount - getAmount(index);

    // we can't take away more than we have
    if (diff < 0 && -diff > added[index-1])
        diff = -added[index-1];

    // virtually add it to underlying object
    int32 done = receiver.addOrRemoveCargo(getCargoTypeFromIndex(index), diff, true);
    added[index-1] += done;

    // re-compute cost
    GCost sum;
    for (int i = 1; i <= NUM_TORPS; ++i)
        if (added[i-1])
            sum += getTorp(i).getTorpedo().getCost() * added[i-1];

    if (added[NUM_TORPS])
        sum += getFighterForPlayer(pay.getPlanet().getOwner()).getCost() * added[NUM_TORPS];

    setCost(sum);
}

void
GStarbaseBuildAmmoAction::commit()
{
    GCargoCostTransaction::commit();
    receiver.mustCommit();
}

void
GStarbaseBuildAmmoAction::cancel()
{
    GCargoCostTransaction::cancel();
    receiver.mustCancel();
}

/** Get required tech upgrade.
    \param index type index, [min_index,max_index]
    \return torp tech level that must be acquired to build the
    requested item. 0 if no tech upgrade needed. */
int
GStarbaseBuildAmmoAction::getTechUpgradeNeeded(int index) const
{
    if (index > 0 && index <= NUM_TORPS) {
        int tech = getTorp(index).getTech();
        if (tech > pay.getPlanet().getBaseTech(tl_Torpedo))
            return tech;
    }
    return 0;
}
