/**
  *  \file game/cargo.h
  *  \brief Cargo Management
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CARGO_H
#define PCC_V2_GAME_CARGO_H

#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/transact.h"
#include "game/error.h"
#include "cpluslib/smallset.h"
#include "game/cost.h"

/** Cargo Arbiter. This class arbitrates interleaved / overlapping
    cargo transaction.

    - when you manipulate cargo, obtain a cargo reservation object
      (foo.getCargoArbiter().incUseCount())
    - when you wish to remove cargo from somewhere, first reserve
      the cargo on the cargo handler. It will disappear from display.
    - finally, you can commit the transaction by removing the
      reservation, and performing the pending add and remove operations.
    - drop the reservation object (foo.getCargoArbiter().decUseCount()))

    The cargo arbiter does not itself know how much cargo is
    available, nor does it check any rules. It just manages
    reservations. */
class GCargoArbiter {
    GCargoArbiter(const GCargoArbiter&);
    GCargoArbiter& operator=(const GCargoArbiter&);
    struct Impl;
    Impl* data;
 public:
    GCargoArbiter()
        : data(0) { }
    ~GCargoArbiter();

    void incUseCount();
    void decUseCount() throw();

    /** Return reserved amount, for user interface purposes. */
    int32 getReservedCargo(GCargoType t) const;

    /** Change reservation. */
    int32 changeReservedCargo(GCargoType t, int32 diff) throw();
};

//! Abstract cargo container.
class GCargoContainer : public GTransaction {
 public:
    enum {
        can_Unload     = 1,     ///< this object can provide stuff for unloading
        can_GetUnload  = 2,     ///< this object can receive unloaded stuff
        can_Sup2mc     = 4,     ///< this object can sell supplies

        can_Overload   = 8      ///< do overload. Set by driver code.
    };
 private:
    int32   diffs[el_COUNT];    ///< Target differences.
    int32   reserved[el_COUNT]; ///< Amounts reserved for us. Probably not needed.
    int     flags;
 protected:
    // Interface to underlying object
    /** Return amount this object has (reserved stuff has been subtracted). */
    virtual int32 getCargoOnObject(GCargoType type) = 0;
    /** Add some amount to object (may be negative). */
    virtual void changeCargoOnObject(GCargoType type, int32 amount) = 0;
    /** Reserve some amount on object (negative = cancel reservation). */
    virtual void changeReservedCargo(GCargoType type, int32 amount) = 0;
 public:
    GCargoContainer();
    virtual ~GCargoContainer();

    // Public interface to underlying object
    /** Check whether this object can carry the specified cargo type. */
    virtual bool canHaveCargo(GCargoType cargo_id) = 0;
    /** Get maximum amount of cargo this object can carry. */
    virtual int32 getMaxCargo(GCargoType cargo_id) = 0;
    /** Get name of underlying object. Used for presenting this object
        to the user. */
    virtual string_t getName() const = 0;

    // Public manipulation interface
    int32 getCargo(GCargoType cargo_id);
    int32 removeCargo(GCargoType type, int32 amount, bool partial);
    int32 addCargo(GCargoType type, int32 amount, bool partial);
    int32 addOrRemoveCargo(GCargoType type, int32 amount, bool partial);

    /** Get container properties. */
    int getFlags() const
        { return flags; }
    /** Set container properties. */
    void setFlags(int fl)
        { flags = fl; }
    /** Check whether container has a certain property. */
    bool hasFlag(int fl) const
        { return flags & fl; }

 protected:
    // Transaction interface
    void cancel();
    void commit();
};

//! Cargo transfer transaction.
class GCargoTransfer : public GTransaction {
    GCargoContainer* objs[2];
 public:
    GCargoTransfer(GCargoContainer& left, GCargoContainer& right);
    ~GCargoTransfer();

    /** Get object on side.
        \param side which object you want, 0 or 1 */
    GCargoContainer& getContainer(int side)
        { return *objs[side]; }

    int32 move(GCargoType type, int32 amount, int from, bool partial, bool s2m);

    bool doUnload(bool s2m);

 protected:
    void cancel();
    void commit();
};

//! Transaction which costs some cargo.
class GCargoCostTransaction : public GTransaction {
    enum { LIMIT = el_Torps };

    GCargoContainer& container;

    int32 cost_amounts[LIMIT];
    int32 reserved_amounts[LIMIT];

    bool isValidIndex(GCargoType t) const;

 protected:
    void changeCost(GCargoType type, int32 amount);
    void changeCost(const GCost& cost);
    void setCost(GCargoType type, int32 amount);
    void setCost(const GCost& cost);
    void clearCosts();

    void commit();
    void cancel();

 public:
    GCargoCostTransaction(GCargoContainer& container);
    ~GCargoCostTransaction();

    bool isValidCost() const;

    int32 getCost(GCargoType type) const;
    int32 getRemaining(GCargoType type) const;
    int32 getMissing(GCargoType type) const;
    int32 getReserved(GCargoType type) const;

    GError getError();

    GCargoContainer& getContainer() const
        { return container; }
};

#endif
