/**
  *  \file game/host.h
  *  \brief Knowledge about Host Properties
  *
  *  This module contains all host-specific knowledge, except for
  *  actual formulas. Instead of a 'host.isPHost()' switch, every
  *  such  decision should have a function in this module. If a
  *  new difference is found, it should end up here, too.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_HOST_H
#define PCC_V2_GAME_HOST_H

#include "game/config.h"
#include "cpluslib/string.h"

/** Host Type. */
enum GHostType {
    ht_Unknown,                 ///< Not yet determined (must be first).
    ht_Host,                    ///< Tim-Host, any version.
    ht_SRace,                   ///< SRace, any version.
    ht_PHost                    ///< PHost, any version.
};

/** Make version code from three numbers. For example,
    MKVERSION(3,22,31) yiels 322031, the version code for
    THost 3.22.031. If the final component is a letter,
    use its position in the alphabet, e.g. MKVERSION(3,2,5)
    for 3.2e. */
#define MKVERSION(x,y,z) ((x)*100000+(y)*1000+(z))

/** Knowledge about Host properties. This structure provides a
    namespace for all host-specific knowledge required in our client. */
struct GHost {
    /** Host type. */
    GHostType host_type;
    /** Host version. Encoded in the same way as it's visible to scripts:
        322040 = 3.22.40. @see MKVERSION(). */
    long host_version;
    /** Size of HConfig. If we don't get a better information, we use this
        to estimate the host version. */
    int hconfig_size;

    /** Alliance support level for THost. */
    enum TimAllianceSupportLevel {
        ta_NoAlliances,              ///< No alliance support at all.
        ta_SimpleAlliances,          ///< "ff" alliances only.
        ta_StrongAlliances           ///< "FF" alliances (two levels).
    };

    /** Declare that we have a HConfig. */
    inline void setHaveHConfig(int size)
        {
            if (!host_type)
                host_type = ht_Host;
            hconfig_size = size;
        }
    /** Declare that we have a PConfig. */
    inline void setHavePConfig()
        {
            if (!host_type)
                host_type = ht_PHost;
        }

    /** Set host version. Use this only when you know the version
        exactly; from a message maybe. */
    inline void setHostVersion(long v)
        { host_version = v; }

    /** Set host version and type. */
    inline void setHostVersion(long v, GHostType type)
        { host_version = v; host_type = type; }

    /** Check whether this host is the 'PHost' family. */
    inline bool isPHost()
        { return host_type == ht_PHost; }

    /** Check whether this host is the 'Tim-Host' family. SRace is
        reasonably similar to Tim-Host so it is returned as
        Tim-Host. */
    inline bool isTHost()
        { return host_type != ht_PHost; }

    /** Check whether this host is SRace. */
    inline bool isSRace()
        { return host_type == ht_SRace; }

    void finish();
    void clear();

    /** Get ship command argument limit. Many versions have this
        limited to 500, PHost 3.3b raised it to 10000. */
    inline long getCommandArgumentLimit()
        {
            if (isPHost() && host_version >= MKVERSION(3,3,2))
                return 10000;
            else
                return 500;
        }

    /** Check whether this host version has Death Rays. */
    inline bool hasDeathRays()
        { return (isPHost() && host_version >= MKVERSION(4,0,0)); }

    /** Check whether this host has ship-specific hull functions. */
    inline bool hasShipSpecificFunctions()
        { return (isPHost() && host_version >= MKVERSION(4,0,0)); }

    /** Check whether hullfunc.txt assignments are cumulative in this
        host version. This applies to PHost 4.0i/3.4k and later. */
    inline bool hasCumulativeHullfunc()
        {
            return (isPHost()
                    && (host_version >= MKVERSION(4,0,9)
                        || (host_version < MKVERSION(4,0,0)
                            && host_version >= MKVERSION(3,4,11))));
        }

    /** Check whether ImperialAssault implies PlanetImmunity ability.
        This applies to all hosts except for PHost 4.0i and higher,
        where PlanetImmunity is a separate ability. */
    inline bool hasImmuneAssaultShip()
        { return (!isPHost() || host_version < MKVERSION(4,0,9)); }

    /** Check whether this host has restrictions in loading high-tech
        torps onto low-tech bases. This is a bug introduced in
        3.22.031, and - given the time Tim spends on Host 3.x - is
        expected to stay for a while. These new hosts do not allow
        you to unload torpedoes onto a starbase if the base doesn't have
        enough tech to build these torps. */
    inline bool hasHighTechTorpedoBug()
        { return (isTHost() && host_version >= MKVERSION(3,22,31)); }

    /** Check whether siliconoid natives have desert advantage in this
        host. This applies to Tim-Host, as well as PHost 3.3c+. */
    inline bool hasSiliconoidDesertAdvantage()
        { return (isTHost() || (isPHost() && host_version >= MKVERSION(3,3,3))); }

    /** Check whether this host allows large cargo transfers. A large
        cargo transfer is one where the cargo amount in transit plus
        the ship content exceed the cargo room size. This is allowed
        by most Host versions, but was blocked/made configurable in
        recent THosts. */
    inline bool hasLargeCargoTransfer()
        { return (isPHost() || (isTHost() && host_version <= MKVERSION(3,22,30))); }

    /** Check THost alliance support. This function is only applicable
        to Tim's host, and returns the support level.
        - 0 if no alliances supported
        - 1 if normal 'ff' allies
        - 2 if strong 'FF' allies */
    inline TimAllianceSupportLevel getTimAllianceLevel()
        {
            // assume best possible level if host version not known
            if (host_version == 0 || host_version >= MKVERSION(3,22,39))
                return ta_StrongAlliances;
            else if (host_version >= MKVERSION(3,22,7))
                return ta_SimpleAlliances;
            else
                return ta_NoAlliances;
        }

    /** Check whether the "Lay mines in" mission automatically fills
        in the minefield owner. When enlarging a foreing minefield,
        some PHost versions require you to set an 'miX' friendly code,
        some do not. */
    inline bool hasAutomaticMineIdentity()
        { return (isPHost() && host_version >= MKVERSION(3,4,3)); }

    /** Get post-taxation happiness limit. This is the highest level
        of happiness \em{after taxation happiness effect} at which
        you'll still collect taxes */
    inline int getPostTaxationHappinessLimit()
        { return isPHost() ? 30 : 31; }

    /** Check whether host allows negative numeric friendly codes.
        This applies to PHost 2.9 and later. */
    inline bool hasNegativeFCodes()
        { return isPHost() && host_version >= MKVERSION(2,9,0); }

    /** Check whether host allows space-padding in numeric friendly codes.
        This applies to PHost 4.0h or 3.4j and later. */
    inline bool hasSpacePaddedFCodes()
        { return isPHost() &&
              (host_version >= MKVERSION(4,0,8)
               || (host_version < MKVERSION(4,0,0)
                   && host_version >= MKVERSION(3,4,10))); }

    /** Get the maximum tax for this race. In THost, natives cannot be
        taxed above 20. */
    inline int getNativeTaxRateLimit(int player)
        {
            if (isTHost() && raceId(player) == 6)
                return 20;
            else
                return 100;
        }

    /** Check whether the build system of this host has PBP style. */
    inline int isPBPGame()
        {
            return isTHost()
                || strCaseCmp(config.BuildQueue().substr(0, 3), "pbp") == 0;
        }
};
extern GHost host;

#endif
