/**
  *  \file game/ufo.cc
  *  \brief Ufo and Wormhole handling
  *
  *  Ufo loading mechanism: the most important thing Ufos are used for
  *  are wormholes. We handle these specially. Other than that,
  *  UFO.HST Ufos and PHost's General Objects (GOs) all end up here.
  *  We identify an object by its Id/Type-Code combination as we
  *  cannot assume that all add-ons that generate GOs coordinate on an
  *  Id range. As a special exception, Ufos with an Id in the range of
  *  1..1000 need not have the same type code. We do not support
  *  multiple instances with identical Id/Type. (It seems PCC 1.x
  *  cannot handle multiple ufos with the same Id but different type;
  *  but it was always meant to be handled like described here.)
  *
  *  For Ufos and GOs, merging is trivial. We assume both data sources
  *  contain equivalent information; the last seen instance survives.
  *
  *  For wormholes, we have three possible sources:
  *  - the Ufo from KORE.DAT
  *  - the UTIL.DAT entry
  *  - the WORMHOLE.TXT file
  *  Unfortunately, there is no 1:1 mapping between wormhole Ids and
  *  Ufo Ids (each WH consumes two WH Id slots, but whereas a
  *  bidirectional WH consumes two Ufo slots, an unidirectional one
  *  consumes only one). We're indexing by Ufo Id, so the KORE.DAT
  *  information should be processed first. UTIL.DAT entries are
  *  merged later, and must match the sequence. That is, if KORE.DAT
  *  contains Ufos 51,53,54, and UTIL.DAT contains Ufos 0,6,7, they're
  *  matched against each other in this sequence. Finally, WORMHOLE.TXT
  *  is used to "fill in the blanks", mainly intended to be used in
  *  games where wormholes are static and known to everyone in the game.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ufo.h"
#include "game/struct.h"
#include "game/utildata.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "cpluslib/misc.h"

GUfo::GUfo(int id)
    : id(id),
      vga_color(0),
      pos(0, 0),
      warp(0),
      heading(-1),
      planet_range(0), ship_range(0),
      radius(0),
      type_code(0),
      name(),
      info1(), info2(),
      real_id(0),
      turn_last_seen(0),
      pos_last_seen(0, 0),
      movement_vector(0, 0),
      flags(0),
      target(0, 0)
{ }

GUfo::GUfo(const GUfo& other)
    : GCircularObject(),
      id(other.id),
      vga_color(other.vga_color),
      pos(other.pos),
      warp(other.warp),
      heading(other.heading),
      planet_range(other.planet_range),
      ship_range(other.ship_range),
      radius(other.radius),
      type_code(other.type_code),
      name(other.name),
      info1(other.info1),
      info2(other.info2),
      real_id(other.real_id),
      turn_last_seen(other.turn_last_seen),
      pos_last_seen(other.pos_last_seen),
      movement_vector(other.movement_vector),
      flags(other.flags),
      target(other.target)
{ }

GUfo::~GUfo()
{ }

GUfo&
GUfo::operator=(const GUfo& rhs)
{
    id              = rhs.id;
    vga_color       = rhs.vga_color;
    pos             = rhs.pos;
    warp            = rhs.warp;
    heading         = rhs.heading;
    planet_range    = rhs.planet_range;
    ship_range      = rhs.ship_range;
    radius          = rhs.radius;
    type_code       = rhs.type_code;
    name            = rhs.name;
    info1           = rhs.info1;
    info2           = rhs.info2;
    real_id         = rhs.real_id;
    turn_last_seen  = rhs.turn_last_seen;
    pos_last_seen   = rhs.pos_last_seen;
    movement_vector = rhs.movement_vector;
    flags           = rhs.flags;
    target          = rhs.target;
    return *this;
}

/* GCircularObject interface */
string_t
GUfo::getName() const
{
    return name;
}

string_t
GUfo::getFullName() const
{
    return format("%s (#%d)", name, id);
}

int
GUfo::getId() const
{
    return id;
}

int
GUfo::getOwner() const
{
    return 0;
}

bool
GUfo::isExistant() const
{
    return vga_color != 0;
}

GPoint
GUfo::getPos() const
{
    return pos;
}

int
GUfo::getRadius() const
{
    return radius;
}

/* Loader interface */
void
GUfo::addUfoData(const TUfo& data)
{
    copyUfo(data);
    flags |= UfoSeenThisTurn;
}

void
GUfo::addChartData(const TDbUfo& data)
{
    ASSERT(data.id == id);
    copyUfo(data.ufo);
    real_id         = data.real_id;
    turn_last_seen  = data.turn_last_seen;
    pos_last_seen   = GPoint(data.x_last_seen, data.y_last_seen);
    movement_vector = GPoint(data.x_move, data.y_move);
    flags          |= UfoStoredInDatabase;
}

void
GUfo::addWormholeData(const TUtil14Wormhole& /*data*/)
{
    // FIXME (xref ccmain.pas::ConsumeWormhole)
}

void
GUfo::copyUfo(const TUfo& data)
{
    vga_color    = data.color;
    name         = getBasicString(data.name);
    info1        = getBasicString(data.info1);
    info2        = getBasicString(data.info2);
    pos          = GPoint(data.x, data.y);
    warp         = data.warp;
    heading      = data.heading;
    planet_range = data.planet_range;
    ship_range   = data.ship_range;
    radius       = data.radius;
    type_code    = data.type_code;
}

int
GUfo::getColorCode() const
{
    return vga_color;
}

uint8
GUfo::getColor() const
{
    static const uint8 vga_to_ours[] = {
        COLOR_BLACK,     COLOR_DARKBLUE,      COLOR_DARKGREEN,    COLOR_DARKCYAN,
        COLOR_DARKRED,   COLOR_DARKMAGENTA,   COLOR_DARKYELLOW,   COLOR_GRAY,
        COLOR_DARK,      COLOR_BRIGHTBLUE,    COLOR_BRIGHTGREEN,  COLOR_BRIGHTCYAN,
        COLOR_BRIGHTRED, COLOR_BRIGHTMAGENTA, COLOR_BRIGHTYELLOW, COLOR_WHITE
    };
    if (vga_color >= 0 && vga_color < countof(vga_to_ours))
        return vga_to_ours[vga_color];
    else
        return COLOR_WHITE;
}

int
GUfo::getSpeed() const
{
    return warp;
}

int
GUfo::getHeading() const
{
    return heading;
}

/***************************** GUfoContainer *****************************/


GUfoContainer::GUfoContainer()
{ }

GUfoContainer::~GUfoContainer()
{ }

GUfo&
GUfoContainer::getUfoById(int id)
{
    iterator i = ufos.begin();
    while (i != ufos.end() && i->getId() < id)
        ++i;

    if (i == ufos.end()) {
        /* does not exist, and has higher Id than we know so far */
        ufos.push_back(GUfo(id));
        return ufos.back();
    } else if (i->getId() == id) {
        /* does already exist */
        return *i;
    } else {
        /* does not exist, have to insert in the middle */
        return *ufos.insert(i, GUfo(id));
    }
}
