/**
  *  \file ui/iconbox.cc
  *  \brief Icon box widget
  */

#include "ui/iconbox.h"
#include "gfx/clip.h"
#include "ui/skin.h"

/** Constructor.
    \param size Iconbox size */
UIIconBox::UIIconBox(GfxPoint size)
    : size(size),
      current_item(0),
      hover_item(-1),
      left_x(0),
      pending_scroll(false)
{ }

void
UIIconBox::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = size;
    info.flags = li_FlexibleH;
}

void
UIIconBox::draw(GfxCanvas& can)
{
    int count = getNumberOfItems();
    int x = 0;
    for (int i = 0; i < count; ++i) {
        int wi = getItemWidth(i);
        int itemx = extent.x - left_x + x;
        int itemy = extent.y;
        GfxRect cliprect(itemx, itemy, wi, extent.h);
        cliprect.intersect(extent);
        if (cliprect.exists()) {
            GfxClipRectFilter filter(cliprect, can);
            drawItem(i, filter, GfxRect(itemx, itemy, wi, extent.h),
                     i == current_item ? Selected : i == hover_item ? Hover : Normal);
        }
        x += wi;
    }

    int curx = extent.x - left_x + x;
    if (curx < extent.getRightX())
        getSkin().drawBackground(can, GfxRect(curx, extent.y, extent.getRightX() - curx, extent.h));

    can.update(extent);
}

void
UIIconBox::stateChanged(int state, bool enable)
{
    if ((state & st_Selected) != 0 && !enable) {
        if (hover_item > 0) {
            hover_item = -1;
            drawWidget();
        }
        if (pending_scroll) {
            if (adjustPosition())
                drawWidget();
            pending_scroll = false;
        }
    }
    UIWidget::stateChanged(state, enable);
}


bool
UIIconBox::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        if (extent.contains(e.mouse.x, e.mouse.y)) {
            /* mouse points into this widget */
            select();

            /* find item mouse points to */
            int x = 0;
            int new_item = -1;
            int count = getNumberOfItems();
            for (int i = 0; i < count; ++i) {
                int wi = getItemWidth(i);
                if (e.mouse.x >= extent.x - left_x + x && e.mouse.x < extent.x - left_x + x + wi) {
                    new_item = i;
                    break;
                }
                x += wi;
            }

            bool redraw = false;
            if (e.mouse.b != 0) {
                /* mouse button is pressed, so select that item (but don't scroll,
                   that would immediately select another item) */
                if (hover_item >= 0) {
                    hover_item = -1;
                    redraw = true;
                }
                if (new_item != current_item && new_item >= 0) {
                    current_item = new_item;
                    sig_changed.raise(current_item);
                    redraw = true;
                    pending_scroll = true;
                }
            } else {
                /* mouse not pressed, just hover */
                if (new_item != hover_item) {
                    hover_item = new_item;
                    redraw = true;
                }
                if (pending_scroll) {
                    /* we've selected a different item during last mouse-click, so
                       we need to scroll */
                    if (adjustPosition())
                        redraw = true;
                    pending_scroll = false;
                }
            }
            if (redraw) {
                drawWidget();
            }
            return true;
        } else {
            if (hover_item >= 0) {
                hover_item = -1;
                drawWidget();
            }
            return UIWidget::handleEvent(e, second_pass);
        }
    } else {
        return UIWidget::handleEvent(e, second_pass);
    }
}

/** Set current item. Also adjusts display so that the item is actually visible.
    \param n Item index [0,getNumberOfItems()). */
void
UIIconBox::setCurrent(int n)
{
    bool redraw_needed = false;
    bool signal_needed = false;

    int total = getNumberOfItems();
    if (total > 0) {
        /* range check */
        if (n >= total)
            n = total-1;
        if (n < 0)
            n = 0;

        /* select the item */
        if (current_item != n) {
            current_item = n;
            signal_needed = true;
            redraw_needed = true;
        }

        if (adjustPosition())
            redraw_needed = true;
    } else {
        /* degenerate case */
        if (current_item > 0 || left_x > 0)
            redraw_needed = true;
        current_item = 0;
        left_x = 0;
    }

    /* redraw? */
    if (redraw_needed)
        drawWidget();
    if (signal_needed)
        sig_changed.raise(current_item);
}

/** Adjust position so that current item is completely visible. */
bool
UIIconBox::adjustPosition()
{
    /* make sure item is visible */
    int this_item_width = getItemWidth(current_item);
    int this_item_x = 0;
    for (int i = 0; i < current_item; ++i)
        this_item_x += getItemWidth(i);

    int new_left_x = left_x;

    if (this_item_x < left_x) {
        /* item is to the left of visible area */
        new_left_x = this_item_x - 10;
        if (new_left_x < 0)
            new_left_x = 0;
    } else if (this_item_x + this_item_width > left_x + extent.w) {
        /* item is to the right of visible area */
        new_left_x = this_item_x + this_item_width - extent.w + 10;
        if (new_left_x < 0)
            new_left_x = 0;
    }

    /* commit change */
    pending_scroll = false;
    if (new_left_x != left_x) {
        left_x = new_left_x;
        return true;
    } else {
        return false;
    }
}

